import concurrent.futures
from datetime import datetime
from getopt import getopt
import logging
import time

import cache_requests
from lxml import html
import punters_client
import pymongo
import racing_data
import redis
import requests

from .date_utils import *
from .profiling_utils import *


class Command:
    """Common functionality for command line utilities"""

    @classmethod
    def main(cls, args):
        """Main entry point for console script"""

        config = cls.parse_args(args)
        command = cls(**config)
        log_time('processing dates from {date_from:%Y-%m-%d} to {date_to:%Y-%m-%d}'.format(date_from=config['date_from'], date_to=config['date_to']), command.process_dates, config['date_from'], config['date_to'])

    @classmethod
    def parse_args(cls, args):
        """Return a dictionary of configuration values based on the provided command line arguments"""
        
        config = {
            'backup_database':  False,
            'database_uri':     'mongodb://localhost:27017/predictive_punter',
            'date_from':        datetime.now(),
            'date_to':          datetime.now(),
            'logging_level':    logging.INFO,
            'redis_uri':        'redis://localhost:6379/predictive_punter'
        }

        opts, args = getopt(args, 'bd:qr:v', ['backup-database', 'database-uri=', 'quiet', 'redis-uri=', 'verbose'])

        for opt, arg in opts:

            if opt in ('-b', '--backup-database'):
                config['backup_database'] = True

            elif opt in ('-d', '--database-uri'):
                config['database_uri'] = arg

            elif opt in ('-q', '--quiet'):
                config['logging_level'] = logging.WARNING

            elif opt in ('-r', '--redis-uri'):
                config['redis_uri'] = arg

            elif opt in ('-v', '--verbose'):
                config['logging_level'] = logging.DEBUG

        if len(args) > 0:
            config['date_from'] = config['date_to'] = datetime.strptime(args[-1], '%Y-%m-%d')
            if len(args) > 1:
                config['date_from'] = datetime.strptime(args[0], '%Y-%m-%d')

        return config

    def __init__(self, *args, **kwargs):

        logging.basicConfig(level=kwargs['logging_level'])

        database_client = pymongo.MongoClient(kwargs['database_uri'])
        self.database = database_client.get_default_database()
        self.do_database_backups = kwargs['backup_database']

        http_client = None
        try:
            http_client = cache_requests.Session(connection=redis.fromurl(kwargs['redis_uri']))
        except BaseException:
            try:
                http_client = cache_requests.Session()
            except BaseException:
                http_client = requests.Session()

        html_parser = html.fromstring

        scraper = punters_client.Scraper(http_client, html_parser)
        
        self.provider = racing_data.Provider(self.database, scraper)

    def backup_database(self):
        """Backup the database if backup_database is available"""

        for backup_name in [collection for collection in self.database.collection_names(False) if '_backup' in collection]:
            self.database.drop_collection(backup_name)

        for collection_name in [collection for collection in self.database.collection_names(False) if '_backup' not in collection]:
            backup_name = collection_name + '_backup'
            self.database[collection_name].aggregate([{'$out': backup_name}])

    def restore_database(self):
        """Restore the database if backup_database is available"""

        for collection_name in [collection for collection in self.database.collection_names(False) if '_backup' not in collection]:
            self.database.drop_collection(collection_name)

        for backup_name in [collection for collection in self.database.collection_names(False) if '_backup' in collection]:
            collection_name = backup_name.replace('_backup', '')
            self.database[backup_name].aggregate([{'$out': collection_name}])

    def process_collection(self, collection, target):
        """Asynchronously process all items in collection via target"""

        if len(collection) > 0:

            with concurrent.futures.ThreadPoolExecutor() as executor:

                def process_item(item, retry_count=0, max_retries=5):
                    try:
                        return executor.submit(log_time, 'processing {item}'.format(item=item), target, item)
                    except RuntimeError:
                        if retry_count < max_retries:
                            time.sleep(1)
                            return process_item(item, retry_count + 1, max_retries)
                        else:
                            raise

                futures = [process_item(item) for item in collection]

                for future in concurrent.futures.as_completed(futures):
                    if future.exception() is not None:
                        raise future.exception()

    def process_dates(self, date_from, date_to):
        """Process all racing data for the specified date range"""

        for date in dates(date_from, date_to):
            log_time('processing date {date:%Y-%m-%d}'.format(date=date), self.process_date, date)

    def process_date(self, date):
        """Process all racing data for the specified date"""

        try:
            self.process_collection(self.provider.get_meets_by_date(date), self.process_meet)

        except BaseException:
            logging.critical('An exception occurred while processing date {date:%Y-%m-%d}'.format(date=date))
            if self.do_database_backups:
                log_time('restoring database from backup', self.restore_database)
            raise

        else:
            if self.do_database_backups:
                log_time('backing up the database', self.backup_database)

    def process_meet(self, meet):
        """Process the specified meet"""

        self.process_collection(meet.races, self.process_race)

    def process_race(self, race):
        """Process the specified race"""

        self.process_collection(race.runners, self.process_runner)

    def process_runner(self, runner):
        """Process the specified runner"""

        if runner.horse is not None:
            log_time('processing {horse}'.format(horse=runner.horse), self.process_horse, runner.horse)

        if runner.jockey is not None:
            log_time('processing {jockey}'.format(jockey=runner.jockey), self.process_jockey, runner.jockey)

        if runner.trainer is not None:
            log_time('processing {trainer}'.format(trainer=runner.trainer), self.process_trainer, runner.trainer)

    def process_horse(self, horse):
        """Process the specified horse"""

        horse.performances

    def process_jockey(self, jockey):
        """Process the specified jockey"""

        pass

    def process_trainer(self, trainer):
        """Process the specified trainer"""

        pass



import predictive_punter
import pymongo
import pytest


@pytest.fixture(scope='module')
def database(database_uri):

    database_name = database_uri.split('/')[-1]
    database_client = pymongo.MongoClient(database_uri)
    database_client.drop_database(database_name)
    return database_client.get_default_database()


@pytest.fixture(scope='module')
def database_uri():

    return 'mongodb://localhost:27017/predictive_punter_test'


@pytest.fixture(scope='module')
def scrape_command(database_uri):

    predictive_punter.ScrapeCommand.main(['-b', '-d', database_uri, '2016-2-1', '2016-2-2'])


def count_distinct(collection, key, exclude=None):
    """Get the number of distinct values for key in the specified database collection"""

    values = collection.distinct(key)
    if exclude is not None and exclude in values:
        values.remove(exclude)
    return len(values)


def test_meets(database, scrape_command):
    """The scrape command should populate the database with the expected number of meets"""

    assert database['meets'].count() == database['meets_backup'].count() == 5


def test_races(database, scrape_command):
    """The scrape command should populate the database with the expected number of races"""

    assert database['races'].count() == database['races_backup'].count() == 8 + 8 + 8 + 7 + 8


def test_runners(database, scrape_command):
    """The scrape command should populate the database with the expected number of runners"""

    assert database['runners'].count() == database['runners_backup'].count() == 11 + 14 + 14 + 15 + 10 + 17 + 11 + 15 + 8 + 9 + 11 + 9 + 11 + 10 + 13 + 16 + 10 + 6 + 11 + 11 + 9 + 9 + 14 + 10 + 9 + 11 + 13 + 18 + 11 + 10 + 14 + 10 + 15 + 13 + 12 + 11 + 11 + 14 + 16


def test_horses(database, scrape_command):
    """The scrape command should populate the database with the expected number of horses"""

    assert database['horses'].count() == database['horses_backup'].count() == count_distinct(database['runners'], 'horse_url', 'https://www.punters.com.au')


def test_jockeys(database, scrape_command):
    """The scrape command should populate the database with the expected number of jockeys"""

    assert database['jockeys'].count() == database['jockeys_backup'].count() == count_distinct(database['runners'], 'jockey_url', 'https://www.punters.com.au/')


def test_trainers(database, scrape_command):
    """The scrape command should populate the database with the expected number of trainers"""

    assert database['trainers'].count() == database['trainers_backup'].count() == count_distinct(database['runners'], 'trainer_url', 'https://www.punters.com.au/')


def test_performances(database, scrape_command):
    """The scrape command should populate the database with the expected number of performances"""

    assert database['performances'].count() >= database['runners'].count({'is_scratched': False})
    assert database['performances_backup'].count() == database['performances'].count()



from datetime import datetime
import logging


def log_time(message, target, *target_args, **target_kwargs):

    start_time = datetime.now()
    logging.info('Started {message} at {start_time}'.format(message=message, start_time=start_time))

    output = target(*target_args, **target_kwargs)

    elapsed_time = datetime.now() - start_time
    logging.info('Finished {message} in {elapsed_time}'.format(message=message, elapsed_time=elapsed_time))

    return output



import sys

from . import Command


class ScrapeCommand(Command):
    """Command line utility to scrape all racing data for a specified date range"""
    pass


def main():
    """Main entry point for scrape console script"""

    ScrapeCommand.main(sys.argv[1:])
